/*
 * Copyright (c) 2003 Frank Sauer. All rights reserved.
 *
 * Licenced under CPL 1.0 (Common Public License Version 1.0).
 * The licence is available at http://www.eclipse.org/legal/cpl-v10.html.
 *
 *
 * DISCLAIMER OF WARRANTIES AND LIABILITY:
 *
 * THE SOFTWARE IS PROVIDED "AS IS".  THE AUTHOR MAKES  NO REPRESENTATIONS OR WARRANTIES,
 * EITHER EXPRESS OR IMPLIED.  TO THE EXTENT NOT PROHIBITED BY LAW, IN NO EVENT WILL THE
 * AUTHOR  BE LIABLE FOR ANY DAMAGES, INCLUDING WITHOUT LIMITATION, LOST REVENUE,  PROFITS
 * OR DATA, OR FOR SPECIAL, INDIRECT, CONSEQUENTIAL, INCIDENTAL  OR PUNITIVE DAMAGES,
 * HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF  LIABILITY, ARISING OUT OF OR RELATED TO
 * ANY FURNISHING, PRACTICING, MODIFYING OR ANY USE OF THE SOFTWARE, EVEN IF THE AUTHOR
 * HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 *
 * $id$
 */
package net.sourceforge.metrics.core;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import net.sourceforge.metrics.core.sources.Cache;
import net.sourceforge.metrics.propagators.Propagator;
import net.sourceforge.metrics.propagators.Sum;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.ListenerList;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.plugin.AbstractUIPlugin;

/**
 * The main plugin class providing access to the metrics framework.
 * 
 * @author Frank Sauer
 */
public class MetricsPlugin extends AbstractUIPlugin implements IPropertyChangeListener, Constants {

	
	private HashMap metricsDependencies;
	private String[] descriptions;
	private String[] ids;
	//The shared instance.
	private static MetricsPlugin plugin;
	//Resource bundle.
	private ResourceBundle resourceBundle;
	
	private boolean listenerAdded = false;
	
	private HashMap calculators = new HashMap();
	private HashMap metrics = new HashMap();
	private HashMap exporters = new HashMap();
	private ListenerList listeners = new ListenerList();
	
	public static ImageDescriptor createImage(String name) {
		return ImageDescriptor.createFromURL(makeImageURL(name));
	}

	private static URL makeImageURL(String name) {
		String path = "icons/" + name;
		URL url = null;

		try {
			url = new URL(MetricsPlugin.getDefault().getDescriptor().getInstallURL(), path);
		} catch (MalformedURLException e) {
			Log.logError("Image not found", e);

			return null;
		}

		return url;
	}	
	/**
	 * Creates the plugin instance and installs metrics and calculators
	 * specified in the manifest(s).
	 */
	public MetricsPlugin(IPluginDescriptor descriptor) {
		super(descriptor);
		plugin = this;
		try {
			resourceBundle= ResourceBundle.getBundle("net.sourceforge.metrics.core.MetricsPluginResources");
		} catch (MissingResourceException x) {
			resourceBundle = null;
		}
		try {
			installExtensions();
		} catch (Throwable t) {
			t.printStackTrace();
		}
	}

	/**
	 * Returns the shared instance.
	 */
	public static MetricsPlugin getDefault() {
		return plugin;
	}
	
	/**
	 * get a list of all installed metric ids
	 * @return String[]
	 */
	public String[] getMetricIds() {
		if (ids == null) {
			ids = parsePrefString(false);
		}
		return ids;
	}
	
	/**
	 * get a list of all installed metric descriptions
	 * @return String[]
	 */
	public String[] getMetricDescriptions() {
		if (descriptions == null) {
			descriptions = parsePrefString(true);
		}
		return descriptions;
	}
	
	public MetricDescriptor getMetricDescriptor(String id) {
		return (MetricDescriptor) metrics.get(id);
	}
	
	private String[] parsePrefString(boolean description) {
		String stringList = getPreferenceStore().getString("METRICS.displayOrder");
		StringTokenizer t = new StringTokenizer(stringList, ",");
		int length = t.countTokens();
		String[] items = new String[length];
		for (int i = 0; i < length; i++) {
			String next = t.nextToken();
			int dash = next.indexOf('-');
			if (description) {
				items[i] = next.substring(dash+1).trim();
			} else {
				items[i] = next.substring(0,dash).trim();
			}
		}
		return items;
	}
	
	/**
	 * Returns the string from the plugin's resource bundle,
	 * or 'key' if not found.
	 */
	public static String getResourceString(String key) {
		ResourceBundle bundle= MetricsPlugin.getDefault().getResourceBundle();
		try {
			return bundle.getString(key);
		} catch (MissingResourceException e) {
			return key;
		}
	}

	/**
	 * Returns the plugin's resource bundle,
	 */
	public ResourceBundle getResourceBundle() {
		return resourceBundle;
	}
	
	public static long lastTimePreferencesChanged() {
		return getDefault().getPreferenceStore().getLong("METRICS.lastPrefChange");
	}
	
	/**
	 * record the last time the preferences were change to force recalculation
	 * after a change.
	 */
	public static void recordTimeAndClearCache() {
		//System.err.println("Recording preference change timestamp.");
		getDefault().getPreferenceStore().putValue("METRICS.lastPrefChange", String.valueOf(new Date().getTime()));
		//Cache.singleton.clear();
	}
	
	/**
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#getPreferenceStore()
	 */
	public IPreferenceStore getPreferenceStore() {
		if (!listenerAdded) {
			super.getPreferenceStore().setDefault("METRICS.decimals", FRACTION_DIGITS);
			super.getPreferenceStore().setDefault("METRICS.xmlformat", "net.sourceforge.metrics.internal.xml.MetricsFirstExporter");
			super.getPreferenceStore().setDefault("METRICS.enablewarnings", false);
			super.getPreferenceStore().setDefault("METRICS.defaultColor", "0,0,0");
			super.getPreferenceStore().setDefault("METRICS.linkedColor", "0,0,255");
			super.getPreferenceStore().setDefault("METRICS.outOfRangeColor", "255,0,0");
			super.getPreferenceStore().setDefault("METRICS.depGR_background", "1,17,68");
			super.getPreferenceStore().setDefault("METRICS.showProject", true);
			super.getPreferenceStore().addPropertyChangeListener(this);
			listenerAdded = true;
		}
		return super.getPreferenceStore();
	}
	
	public static boolean isWarningsEnabled() {
		return getDefault().getPreferenceStore().getBoolean("METRICS.enablewarnings");
	}
	
	/**
	 * @see org.eclipse.jface.util.IPropertyChangeListener#propertyChange(org.eclipse.jface.util.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent event) {
		if (!event.getProperty().startsWith("METRICS")) {
			recordTimeAndClearCache();
		} else {
			ids = null;
			descriptions = null;
		}
		Object[] l = listeners.getListeners();
		for (int i = 0; i < l.length; i++) {
			if (l[i] instanceof IPropertyChangeListener)
			((IPropertyChangeListener)l[i]).propertyChange(event);
		}
	}

	private void installExtensions() {
		installMetrics();
		installExporters();
	}
	
	/**
	 * scan plugins for metrics and calculators and install them.
	 */
	private void installMetrics() {
		//System.err.println("Discovering and installing metrics");
		IExtensionPoint p = getDescriptor().getExtensionPoint("metrics");
		if (p != null) {
			IExtension[] x = p.getExtensions();
			for (int i = 0; i < x.length;i++) {
				IConfigurationElement[] elements = x[i].getConfigurationElements();
				for (int j = 0; j < elements.length; j++) {
					IConfigurationElement next = elements[j];
					String kind = next.getName();
					if (kind.equals("calculator")) {
						installCalculator(next);
					} else if (kind.equals("metric")) {
						installMetric(next);
					}
				}
			}
			initDisplayOrderPreference();
		}
	}
	
	private void installExporters() {
		//System.err.println("Discovering and installing metrics");
		IExtensionPoint p = getDescriptor().getExtensionPoint("exporters");
		if (p != null) {
			IExtension[] x = p.getExtensions();
			for (int i = 0; i < x.length;i++) {
				IConfigurationElement[] elements = x[i].getConfigurationElements();
				for (int j = 0; j < elements.length; j++) {
					IConfigurationElement next = elements[j];
					String kind = next.getName();
					if (kind.equals("exporter")) {
						installExporter(next);
					} 
				}
			}
		}
	}
	
	private void installExporter(IConfigurationElement element) {
		ExportDescriptor xd = ExportDescriptor.createFrom(element);
		if (xd != null) {
			exporters.put(xd.getClassName(), xd);
		}
	}

	public ExportDescriptor[] getExporters() {
		return (ExportDescriptor[])exporters.values().toArray(new ExportDescriptor[]{});
	}
	
	public ExportDescriptor getExporter(String className) {
		return (ExportDescriptor)exporters.get(className);
	}
	
	public IExporter getCurrentExporter() {
		String format = getPreferenceStore().getString("METRICS.xmlformat");
		return createExporter(format);
	}
	
	public boolean showProjectOnCompletion() {
		return getPreferenceStore().getBoolean("METRICS.showProject");
	}
	
	public IExporter createExporter(String className) {
		ExportDescriptor xd = getExporter(className);
		if (xd != null) {
			return xd.createExporter();
		}
		return null;
	}
	
	/**
	 * Set the default display order.
	 */
	private void initDisplayOrderPreference() {
		StringBuffer list = new StringBuffer();
		for (Iterator i = metrics.keySet().iterator();i.hasNext();) {
			String id = (String)i.next();
			String desc = ((MetricDescriptor)metrics.get(id)).getName();
			list.append(id).append(" - ").append(desc).append(',');
		}
		String def = list.substring(0,list.length()-1);
		getPreferenceStore().setDefault("METRICS.displayOrder", def);
		// if metrics were added/removed, reset the value
		if (getMetricIds().length != metrics.size()) {
			getPreferenceStore().setToDefault("METRICS.displayOrder");
		}
	}
	
	private void installMetric(IConfigurationElement element) {
		MetricDescriptor m = MetricDescriptor.createFrom(element);
		if (m != null) {
			if (m.isSum()) {
				// metric is not really a new metric, but merely the sum
				// of some other metric, make sure it gets calculated
				Sum s = new Sum(m.getId(), m.getSumOf());
				getCalculators(m.getLevel()).add(s);
				//System.err.println("Added a sum: " + s);
			}
			metrics.put(m.getId(), m);
			installPropagators(m);
			recordDependencies(m.getId(), m.getRequiredMetricIds());
		}
	}

	/**
	 * @param string
	 * @param strings
	 */
	private void recordDependencies(String id, String[] requires) {
		if (metricsDependencies == null) {
			metricsDependencies = new HashMap();
		}
		if (requires != null && requires.length > 0) {
			for (int i = 0; i < requires.length; i++) {
				List existing = (List)metricsDependencies.get(requires[i]);
				if (existing == null) {
					existing = new ArrayList();
				}
				existing.add(id);
				metricsDependencies.put(requires[i], existing);
			}
		}
	}

	private void installCalculator(IConfigurationElement nextCalculator) {
		CalculatorDescriptor c = createDescriptorFrom(nextCalculator);
		addCalculator(c);
	}
		
	private void addCalculator(CalculatorDescriptor c) {
		List calcs = getCalculators(c.getLevel());
		calcs.add(c.createCalculator());
		//System.err.println(c);
	}
	
	/**
	 * Install all propagators required by m
	 * @param m
	 */
	private void installPropagators(MetricDescriptor m) {
		String nextLevel = m.getParentLevel();
		if (nextLevel != null) {
			List props = m.createPropagators();
			for (Iterator i = props.iterator(); i.hasNext();) {
				Propagator p = (Propagator)i.next();
				getCalculators(nextLevel).add(p);
				propagatePropagator(nextLevel, p, m);
			}
			String newAvgMaxAt = m.getNewAvgMaxAt();
			if (newAvgMaxAt != null) {
				props = m.createIntroducedAvgMax();
				for (Iterator i = props.iterator(); i.hasNext();) {
					Propagator p = (Propagator)i.next();
					//System.err.println("Installing a " + p + " at level " + newAvgMaxAt);
					getCalculators(newAvgMaxAt).add(p);
					propagatePropagator(newAvgMaxAt, p, m);
				}
			}
		}
	}
	
	/**
	 * Create MaxMax, AvgAvg and Sum calculators at each level above
	 * the given level
	 * @param level
	 * @param p
	 */
	private void propagatePropagator(String level, Propagator p, MetricDescriptor m) {
		String nextLevel = MetricDescriptor.getNextLevel(level);
		Propagator nextP = p.createNextLevel();
		while (nextLevel != null) {
			getCalculators(nextLevel).add(nextP);
			//System.err.println("Added a " + nextP + " at level " + nextLevel);
			nextLevel = MetricDescriptor.getNextLevel(nextLevel);
			nextP = nextP.createNextLevel();
		}
	}
	
	/**
	 * Get the list of calculators for the given level
	 * @param level	as specified in the xml attribute level
	 * @return List
	 */
	public List getCalculators(String level) {
		List result = (List) calculators.get(level);
		if (result == null) {
			result = new ArrayList();
			calculators.put(level, result);
		}
		return result;
	}
	
	/**
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#shutdown()
	 */
	public void shutdown() throws CoreException {
		super.shutdown();
		Cache.singleton.close();
	}

	/**
	 * Method createCalculatorDescriptorFrom.
	 * @param iConfigurationElement
	 */
	private CalculatorDescriptor createDescriptorFrom(IConfigurationElement element) {
		return CalculatorDescriptor.createFrom(element);
	}
	
	/**
	 * @param listener
	 */
	public void addPropertyChangeListener(IPropertyChangeListener listener) {
		listeners.add(listener);		
	}

	/**
	 * @param listener
	 */
	public void removePropertyChangeListener(IPropertyChangeListener listener) {
		listeners.remove(listener);		
	}

	/**
	 * returns a list of metric ids that depend on the given metric being available
	 * @param descriptor
	 * @return array of metric ids or null if none depend on the given metric
	 */
	public String[] getDependentMetrics(MetricDescriptor descriptor) {
		if (descriptor == null) return null;
		List dependents = (List)metricsDependencies.get(descriptor.getId());
		if (dependents != null) return (String[]) dependents.toArray(new String[]{});
		return null;
	}
	
}
